MODULE dksolve
  IMPLICIT NONE
  PRIVATE :: horner, rhorner, userpoly, &
       start, cnvex, cmerge, left, right, ctest
  PUBLIC  :: dksolve
  INTEGER, PARAMETER     :: dp = SELECTED_REAL_KIND(15, 60)
  REAL, PARAMETER        :: eps=1.d-16    ! used for stopping iterations
  LOGICAL, PARAMETER     :: dolog=.false. ! set true for debugging
  real (kind=dp) :: npe ! number of polynomial evaluations  
CONTAINS
!**************************************************************************
! All the software  contained in this library  is protected by copyright. *
! Permission  to use, copy, modify, and  distribute this software for any *
! purpose without fee is hereby granted, provided that this entire notice *
! is included  in all copies  of any software which is or includes a copy *
! or modification  of this software  and in all copies  of the supporting *
! documentation for such software.                                        *
!**************************************************************************
!**************************************************************************
! THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED *
! WARRANTY. IN NO EVENT, THE AUTHORS,  BE  LIABLE FOR ANY ERROR  IN THE   *
! SOFTWARE, ANY MISUSE  OF IT  OR ANY DAMAGE ARISING OUT OF ITS USE. THE  *
! ENTIRE RISK OF USING THE SOFTWARE LIES WITH THE PARTY DOING SO.         *
!**************************************************************************
! ANY USE  OF THE SOFTWARE  CONSTITUTES  ACCEPTANCE  OF THE TERMS  OF THE *
! ABOVE STATEMENT.                                                        *
!**************************************************************************
!
!   REFERENCE AUTHOR:
!
!       DARIO ANDREA BINI
!       UNIVERSITY OF PISA, ITALY
!       E-MAIL: bini@dm.unipi.it
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                        DKSOLVE                            !!!
!!!       APPROXIMATION OF POLYNOMIAL ROOTS BY MEANS OF       !!!
!!!                 DURAND-KERNER ITERATION                   !!!
!!!                         v 0.1 by                          !!!
!!!            D.A. Bini                                      !!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! The program implements the algorithm of Durand-Kerner for the !
! approximation of polynomial zeros.                            !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! SUBROUTINES:
!       DKSOLVE: Implements the algorithm of Durand-Kerner
!       HORNER: Horner's rule for computing p(x)
!       RHORNER: Horner rule for computing x^np(x^(-1))
!       USERPOLY: user's polynomial
!--- The next subroutines, for the choice of starting 
!    approximations, are taken from the package POLZEROS
!    by D. A. Bini, see NUMERICAL COMPUTATION OF POLYNOMIAL ZEROS 
!    BY MEANS OF ABERTH'S METHOD, NUMERICAL ALGORITHMS, 13 (1996), 
!    179-200, adjusted to Fortran90 by Alan Miller, CSIRO 
!    Mathematical & Information Sciences, Private Bag 10, Clayton 
!    South MDC, Victoria, Australia 3169. 
!    Alan.Miller@mel.dms.csiro.au,
!    http://www.mel.dms.csiro.au/~alan
!
!       START
!       CNVX
!       LEFT
!       RIGHT
!       CMERGE
!       CTEST
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                      SUBROUTINE DKSOLVE                       !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Approximates the zeros of the polynomial                      !
!           p(x)=p_0 + p_1*x + ... + p_n*x^n                    !
! by means of an algorithm based on Durand-Kerner iteration:    !
! according to the following steps:                             !
! 1) chose starting approximations z(i), i=1,...,n:             !
!    if the polynomial is assigned in terms of its coefficients !
!    then use the Newton's polygon technique of Bini, Numer.    !
!    Algo. 13, 1996, 179-200. If the polynomial is assigned by  !
!    meas of a subroutine (user polynomial) choose the roots of !
!    the unity shifted with a random rotation.                  !
! 2) compute the Durand-Kerner corrections                      !
!    w(i)=p(z(i))/prod_{j=1,n, j/=i}(z(i)-z(j)), and set        !
!    z(i)=z(i)-w(i), for i=1,...,n.                             !
! 4) repeat from step 2 until convergence.                      !
! The convergence criterion is the one of D. Bini NUMERICAL     !
! COMPUTATION OF POLYNOMIAL ZEROS BY MEANS OF ABERTH'S METHOD,  !
! NUMERICAL ALGORITHMS, 13 (1996), 179-200. This criterion      !
! guarantees that the computed approximation is zero of a       !
! slightly perturbed polynomial.                                !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!  Input variables:                                             !
!  n      : degree of p(x)                                      !
!  p(0:n) : complex coefficients p_i of the polynomial p(x)     !
!  usr    : if usr is true then p(x) is computed by means of    !
!           Horner rule, otherwise the subroutine  userpoly is  !
!           used (user polynomial). In this package, userpoly   !
!           computes the Mandelbrot polynomial defined by       !
!           m_0(x)=1, m_{k+1}(x)=x*m_k(x)^2+1                   !
!  Output variables                                             !
!  z(1:n) : approximations of the zeros                         !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!  Auxiliary variables and parameters:
!  again(1:n): again(i) is false if the approximation z(i) has  !
!              converged                                        !
!  nittot    : number of total<iterations of the power method   !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE dksolve(n, p, usr, z)
    IMPLICIT NONE
    INTEGER                           :: n
    COMPLEX (kind=dp), DIMENSION(0:n) :: p
    COMPLEX (kind=dp), DIMENSION(n)   :: z
    LOGICAL                           :: usr
    !
    INTEGER                         :: nsol, k, i, j, nz, &
         nit, nittot
    LOGICAL, DIMENSION(n)           :: again
    LOGICAL, DIMENSION(n+1)         :: h
    LOGICAL                         :: laux,ag
    REAL (kind=dp)                  :: mx1, mx2, abw
    COMPLEX (kind=dp)               :: tmp, s, zm1
    COMPLEX (kind=dp), DIMENSION(n) :: w
    REAL (kind=dp), DIMENSION(n)    :: rad
    REAL (kind=dp), DIMENSION(0:n)  :: ap
    REAL (kind=dp), DIMENSION(n+1)  :: ap1
    INTEGER, PARAMETER :: nsweeps=5000
    npe=0.d0
    IF(dolog) OPEN(unit=2,file="log.log")
    ! initialize
    nittot=0
    nit=0
    w=0.d0
    nsol=0
    again=.TRUE.
     
    ! choose initial approximations
    h=.TRUE.
    IF(usr)THEN
       DO i=1,n
          z(i)=COS(6.28*i/n+0.223)+SIN(6.28*i/n+0.223)*(0.d0,1.d0)
       END DO
    ELSE
       ap1(1:n+1)=ABS(p(0:n))
       ap(0:n)=ap1(1:n+1)*(eps*n*4.d0)
       CALL  start(n, ap1, z, rad, nz, h)
    END IF
    
    DO k=1,nsweeps
       ! Compute the DK corrections
!       again=.TRUE. !!! ??
       DO i=1,n
       if(.not. again(i))exit
          IF(usr)THEN
             w(i)=0
          ELSE
             w(i)=LOG(p(n))
          END IF
          DO j=1,n
             IF (j.NE.i) THEN
                IF(z(i)-z(j)==0)THEN
                   WRITE(*,*)"WARNING 1 : confluent approximations"
                   STOP
                END IF
                w(i)=w(i)+LOG(z(i)-z(j))
             END IF
          END DO
          IF(usr)THEN
             CALL userpoly(n,z(i),s,again(i))
             w(i)=s-w(i)
          ELSE
             IF(ABS(z(i))<=1) THEN     !! |z(i)|<1
                CALL horner(n,p,ap,z(i),s,again(i))
                IF(s.NE.0)THEN
                   w(i)=LOG(s)-w(i)
                ELSE
                   w(i)=0.d0
                END IF
             ELSE
                zm1=1.d0/z(i)
                CALL rhorner(n,p,ap,zm1,s,again(i))
                IF(s.NE.0)THEN
                   s=LOG(s)+n*LOG(z(i))
                   w(i)=s-w(i)
                ELSE
                   w(i)=0
                END IF
             END IF
          END IF
          w(i)=-EXP(w(i))
!update the approximation
          z(i)=z(i)+w(i)
       END DO
       ! check for stop
       nsol=0
       DO j=1,n
          IF(.NOT. again(j)) THEN
             nsol=nsol+1
          END IF
       END DO
       
       ! sort z with approximated zeros at the end
       DO i=n,2,-1
          IF(again(i)) THEN
             DO j=i-1,1,-1
                IF(.NOT. again(j)) THEN
                   tmp=z(i)
                   z(i)=z(j)
                   z(j)=tmp
                   tmp=w(i)
                   w(i)=w(j)
                   w(j)=tmp
                   again(j)=.TRUE.
                   again(i)=.FALSE.
                   EXIT
                END IF
             END DO
          END IF
       END DO
       ! end of sorting
       
       
       IF(dolog)  THEN
          WRITE(2,*)"sweep ",k-1,"nsol=",nsol
          WRITE(*,*)"sweep ",k-1,"nsol=",nsol
          !do j=1,n
          !   write(4,*)z(j),again(j),abs(w(j))/abs(z(j))
          !end do
       END IF
       
       IF(nsol==n)GOTO 200

    END DO
       

       
200 CONTINUE
    
    IF(dolog) THEN
       WRITE(*,*)"sweep=",k-1," npe=",npe
       WRITE(2,*)"sweep=",k-1," npe=",npe
    END IF
  END SUBROUTINE dksolve
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                      SUBROUTINE HORNER                        !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE horner(n, p, ap, z, s, cnt)
    IMPLICIT NONE
    INTEGER                           :: n
    COMPLEX (kind=dp), DIMENSION(0:n) :: p
    REAL (kind=dp), DIMENSION(0:n)    :: ap
    COMPLEX (kind=dp)                 :: s, z
    LOGICAL                           :: cnt
    !
    INTEGER                           :: i
    REAL (kind=dp)                    :: as, az
    s=p(n)
    as=ap(n)
    az=ABS(z)
    npe=npe+1
    DO i=n-1,0,-1
       s=s*z+p(i)
       as=as*az+ap(i)
    END DO
    cnt=.TRUE.
    IF(ABS(s)<as)cnt=.FALSE.
  END SUBROUTINE horner
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                      SUBROUTINE RHORNER                       !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE rhorner(n, p, ap, z, s, cnt)
    IMPLICIT NONE
    INTEGER                           :: n
    COMPLEX (kind=dp), DIMENSION(0:n) :: p
    REAL (kind=dp), DIMENSION(0:n)    :: ap
    COMPLEX (kind=dp)                 :: s, z
    LOGICAL                           :: cnt
    !
    REAL (kind=dp)                    :: as, az
    INTEGER                           :: i
    npe=npe+1
    s=p(0)
    as=ap(0)
    az=ABS(z)
    DO i=1,n
       s=s*z+p(i)
       as=as*az+ap(i)
    END DO
    cnt=.TRUE.
    IF(ABS(s)<as)cnt=.FALSE.
  END SUBROUTINE rhorner
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                      SUBROUTINE USERPOLY                      !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Compute Log(p(z)) where n=2^k-1 and p(x)=m_k(z) such that     !
!  m_0(z)=1, m_i(z)=z*m_{i-1}^2(z)+1, i=1,2,...,k.              !
!  the union of the zeros of this polynomial sequence generates !
!  the Mandelbrot set.                                          !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  SUBROUTINE userpoly(n, z, w, cnt)
    IMPLICIT NONE
    INTEGER           :: n
    COMPLEX (kind=dp) :: z, w
    LOGICAL           :: cnt
    !
    COMPLEX (kind=dp) :: lw, lz
    REAL (kind=dp)    :: alw, li, alw1, alz
    INTEGER           :: i,k
    real(kind=dp),parameter :: lu=-36.73680057+3, l6=1.791759
    npe=npe+1
    cnt=.TRUE.
    lw=0.d0

    alw=0.d0
    lz=LOG(z)
    alz=real(lz)
    li=-300
    k=LOG(n+1.d0)/LOG(2.d0)+0.5
    DO i=1,k
       lw=lz+lw*2
       IF(real(lw)<40)THEN
          lw=LOG(EXP(lw)-1)
       END IF
    alw1=REAL(lw)
li=-alw1+max(lu,alz+2*alw+max(log(2.d0)+li,log(3.d0)+lu)+log(2.d0))+log(2.d0)
!    li=max(li,lu)+l6+2*alw-alw1+alz
    alw=alw1
    END DO

!    alw1=REAL(lw)
!    lw=lz+lw*2
!    IF(ABS(lw)<37)THEN
!       lw=LOG(EXP(lw)-1)
!    END IF
    !   stop condition
!    IF(alw*2+REAL(lz)-20>REAL(lw))cnt=.FALSE.
   if(li>=1) cnt=.false.
    w=lw
  END SUBROUTINE userpoly
  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      SUBROUTINES FOR INITIAL APPROXIMATIONS                   !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  SUBROUTINE start(n, a, y, radius, nz, h)
    IMPLICIT NONE
    INTEGER                           :: n
    INTEGER                           :: nz
    LOGICAL, dimension(n+1)           :: h  !!!!!!!!!!! n -> n+1
    COMPLEX (KIND(0.d0)),dimension(n) :: y
    REAL (KIND(0.d0))     :: small, big
    REAL (KIND(0.d0)), dimension(n+1) :: a
    REAL (KIND(0.d0)), dimension(n)   :: radius
    ! Local variables
    INTEGER                   :: i, iold, nzeros, j, jj
    REAL (KIND(0.d0))            :: r, th, ang, temp, xsmall, xbig
    REAL (KIND(0.d0)), PARAMETER :: pi2 = 6.2831853071796, sigma = 0.7
    
    SMALL=2.d0**(-1023)
    BIG=2.d0**1023 
    
    xsmall = -1023*log(2.d0) !LOG(small)
    xbig = 1023*log(2.d0)  !LOG(big)
    nz = 0
    ! Compute the logarithm A(I) of the moduli of the coefficients of
    ! the polynomial and then the upper covex hull of the set (A(I),I)
    DO i = 1, n+1
       IF(a(i) /= 0) THEN
          a(i) = LOG(a(i))
       ELSE
          a(i) = -1.d30
       END IF
    END DO
    CALL cnvex(n+1, a, h)
    ! Given the upper convex hull of the set (A(I),I) compute the moduli
    ! of the starting approximations by means of Rouche's theorem
    iold = 1
    th = pi2/n
    DO i = 2, n+1
       IF (h(i)) THEN
          nzeros = i - iold
          temp = (a(iold) - a(i))/nzeros
          ! Check if the modulus is too small
          IF((temp < -xbig).AND.(temp >= xsmall))THEN
             WRITE(*,*)'WARNING:',nzeros,' ZERO(S) ARE TOO SMALL TO'
             WRITE(*,*)'REPRESENT THEIR INVERSES AS COMPLEX (KIND=dp) ::, THEY'
             WRITE(*,*)'ARE REPLACED BY SMALL NUMBERS, THE CORRESPONDING'
             WRITE(*,*)'RADII ARE SET TO -1'
             WRITE(2,*)'WARNING:',nzeros,' ZERO(S) ARE TOO SMALL TO '
             WRITE(2,*)'REPRESENT THEIR INVERSES AS COMPLEX (KIND=dp) ::, THEY'
             WRITE(2,*)'ARE REPLACED BY SMALL NUMBERS, THE CORRESPONDING'
             WRITE(2,*)'RADII ARE SET TO -1'
             nz = nz + nzeros
             r = 1.0D0/big
          END IF
          IF(temp < xsmall)THEN
             nz = nz + nzeros
             WRITE(*,*)'WARNING: ',nzeros,' ZERO(S) ARE TOO SMALL TO BE'
             WRITE(*,*)'REPRESENTED AS COMPLEX (KIND=dp) ::, THEY ARE SET TO 0'
             WRITE(*,*)'THE CORRESPONDING RADII ARE SET TO -1'
             WRITE(2,*)'WARNING: ',nzeros,' ZERO(S) ARE TOO SMALL TO BE'
             WRITE(2,*)'REPRESENTED AS COMPLEX (KIND=dp) ::, THEY ARE SET 0'
             WRITE(2,*)'THE CORRESPONDING RADII ARE SET TO -1'
          END IF
          ! Check if the modulus is too big
          IF(temp > xbig)THEN
             r = big
             nz = nz + nzeros
             WRITE(*,*)'WARNING: ', nzeros, ' ZEROS(S) ARE TOO BIG TO BE'
             WRITE(*,*)'REPRESENTED AS COMPLEX (KIND=dp) ::,'
             WRITE(*,*)'THE CORRESPONDING RADII ARE SET TO -1'
             WRITE(2,*)'WARNING: ',nzeros, ' ZERO(S) ARE TOO BIG TO BE'
             WRITE(2,*)'REPRESENTED AS COMPLEX (KIND=dp) ::,'
             WRITE(2,*)'THE CORRESPONDING RADII ARE SET TO -1'
          END IF
          IF((temp <= xbig).AND.(temp > MAX(-xbig, xsmall)))THEN
             r = EXP(temp)
          END IF
          ! Compute NZEROS approximations equally distributed in the disk of
          ! radius R
          ang = pi2/nzeros
          DO j = iold, i-1
             jj = j-iold+1
             IF((r <= (small)).OR.(r == big)) radius(j) = -1
             y(j) = r*(COS(ang*jj + th*i + sigma) + (0,1)*SIN(ang*jj + th*i + sigma))
          END DO
          iold = i
       END IF
    END DO
    RETURN
  END SUBROUTINE start



  SUBROUTINE cnvex(n, a, h)
    IMPLICIT NONE
    INTEGER, INTENT(IN)        :: n
    LOGICAL, INTENT(OUT)       :: h(:)
    REAL (KIND(0.d0)), INTENT(IN) :: a(:)
    
! Local variables
    INTEGER :: i, j, k, m, nj, jc
    
    h(1:n) = .true.

    ! compute K such that N-2 <= 2**K < N-1
    k = INT(LOG(n-2.0D0)/LOG(2.0D0))
    IF(2**(k+1) <= (n-2)) k = k+1
    
    ! For each M=1,2,4,8,...,2**K, consider the NJ pairs of consecutive
    ! sets made up by M+1 points having the common vertex
    ! (JC,A(JC)), where JC=M*(2*J+1)+1 and J=0,...,NJ,
    ! NJ = MAX(0, INT((N-2-M)/(M+M))).
    ! Compute the upper convex hull of their union by means of subroutine CMERGE
    m = 1
    DO i = 0, k
       nj = MAX(0, INT((n-2-m)/(m+m)))
       DO j = 0, nj
          jc = (j+j+1)*m+1
          CALL cmerge(n, a, jc, m, h)
       END DO
       m = m+m
    END DO
    RETURN
  END SUBROUTINE cnvex


  SUBROUTINE left(h, i, il)
    IMPLICIT NONE
    INTEGER, INTENT(IN)  :: i
    INTEGER, INTENT(OUT) :: il
    LOGICAL, INTENT(IN)  :: h(:)
    
    DO il = i-1, 0, -1
       IF (h(il)) RETURN
    END DO
    RETURN
  END SUBROUTINE left



  SUBROUTINE right(n, h, i, ir)
    IMPLICIT NONE
    INTEGER, INTENT(IN)  :: n, i
    INTEGER, INTENT(OUT) :: ir
    LOGICAL, INTENT(IN)  :: h(:)
    
    DO ir = i+1, n
       IF (h(ir)) RETURN
    END DO
    RETURN
  END SUBROUTINE right
  


  SUBROUTINE cmerge(n, a, i, m, h)
    IMPLICIT NONE
    INTEGER, INTENT(IN)        :: n, m, i
    LOGICAL, INTENT(IN OUT)    :: h(:)
    REAL (KIND(0.d0)), INTENT(IN) :: a(:)
    
    ! Local variables
    INTEGER :: ir, il, irr, ill
    LOGICAL :: tstl, tstr
    
    ! at the left and the right of the common vertex (I,A(I)) determine
    ! the abscissae IL,IR, of the closest vertices of the upper convex
    ! hull of the left and right sets, respectively
    CALL left(h, i, il)
    CALL right(n, h, i, ir)
    
    ! check the convexity of the angle formed by IL,I,IR
    IF (ctest(a, il, i, ir)) THEN
       RETURN
    ELSE
       ! continue the search of a pair of vertices in the left and right
       ! sets which yield the upper convex hull
       h(i) = .false.
       DO
          IF (il == (i-m)) THEN
             tstl = .true.
          ELSE
             CALL left(h, il, ill)
             tstl = ctest(a, ill, il, ir)
          END IF
          IF (ir == MIN(n, i+m)) THEN
             tstr = .true.
          ELSE
             CALL right(n, h, ir, irr)
             tstr = ctest(a, il, ir, irr)
          END IF
          h(il) = tstl
          h(ir) = tstr
          IF (tstl.AND.tstr) RETURN
          IF(.NOT.tstl) il = ill
          IF(.NOT.tstr) ir = irr
       END DO
    END IF
    
    RETURN
  END SUBROUTINE cmerge
  
  


  FUNCTION ctest(a, il, i, ir) RESULT(OK)
    IMPLICIT NONE
    INTEGER, INTENT(IN)        :: i, il, ir
    REAL (KIND(0.d0)), INTENT(IN) :: a(:)
    LOGICAL                    :: OK
    
    ! Local variables
    REAL (KIND(0.d0))            :: s1, s2
    REAL (KIND(0.d0)), PARAMETER :: toler = 0.4D0
    
    s1 = a(i) - a(il)
    s2 = a(ir) - a(i)
    s1 = s1*(ir-i)
    s2 = s2*(i-il)
    OK = .false.
    IF(s1 > (s2+toler)) OK = .true.
    RETURN
  END FUNCTION ctest
  
  
END MODULE dksolve








